RHMHeaders = #("[Geometry]","[Mesh]","[Chunk]")
fn RHMCheckLine fl = (
	result = false
	if not eof fl then (
		stri = readline fl
		for s = 1 to RHMHeaders.count while result == false where RHMHeaders[s] == stri do result = s
	)
	else result  = true
	#(result,stri)
)
fn RHMParseLine stri =if stri != undefined then(
	result = filterstring stri "= ,"
	quote = matchpattern stri pattern:"*\"*\""
	retval = #(result[1],undefined)
	curcnt = 2
	if quote then retval[2] = execute (filterstring stri "=")[2]
	else if result.count > 2 then (
		while result.count > curcnt do (
			token = result[curcnt]
			if token == "(" or token ==")(" then (
				end = undefined;for r = (curcnt+1) to result.count while end == undefined where result[r] ==")" or result[r] ==")(" do end = r
				if end != undefined then (
					itmcnt = end-curcnt-1
					rtitem
					case itmcnt of (
						2: rtitem = [result[curcnt+1] as float, result[curcnt+2] as float]
						3: rtitem = [result[curcnt+1] as float, result[curcnt+2] as float, result[curcnt+3] as float]
						4: rtitem = [result[curcnt+1] as float, result[curcnt+2] as float, result[curcnt+3] as float, result[curcnt+4] as float]
					)
					if retval[2] == undefined then retval[2] = rtitem
					else if (classof retval[2]) != array then retval[2] =#(retval[2], rtitem)
					curcnt = end+ (if result[end] ==")(" then 0 else 1)
				)
			) 
			else curcnt += 1
		)
	)
	else if result.count > 1 then retval[2] = (execute result[2])
	retval
)

fn RHMUniFun rstr fl = (
	cline 
	do (
		cline = RHMCheckLine fl
		if cline[1] == false then (
			pline = RHMParseLine cline[2]
			try (setProperty rstr pline[1] pline[2]) catch print (getCurrentException())
		)
	)while cline[1] == false
	rstr.nextline = cline
)
struct RHMGeometry (
	Meshes,
	Vertices,
	Indices,
	fl,
	nextline,
	on create do if fl !=undefined then RHMUniFun this fl
)

struct RHMMesh (
	Name,
	ChunkStart,
	ChunkCount,
	ChildNum,
	Bounds,
	fl,
	nextline,
	on create do if fl !=undefined then RHMUniFun this fl
)

struct RHMChunk (
	BaseIndex,
	StartIndex,
	PrimCount,
	Bounds,
	Diffuse,
	Specular,
	Emissive,
	Normal,
	DiffuseLayer,
	SpecularLayer,
	NormalLayer,
	vColor,
	fBaseUVTile,
	fSpecularMultiplier,
	fSpecularPowerMultiplier,
	fEnvMultiplier,
	fEmissiveMultiplier,
	fLayerUVTile,
	fWrapAroundTerm,
	material,
	ambient,
	vertexoffset = 0,
	vertexcount,
	varray = #(),
	narray = #(),
	uvarray = #(),
	clarray = #(),
	alarray = #(),
	fl,
	nextline,
	on create do if fl !=undefined then RHMUniFun this fl
)

fn readHalfFloat fstream = (
    hf=readshort fstream #unsigned
    sign = bit.get hf 16
    exponent = (bit.shift (bit.and hf 0x7C00) -10) as integer - 16
    fraction = bit.and hf 0x03FF
    if sign==true then sign = 1 else sign = 0
    exponentF = exponent + 127
    outputAsFloat = bit.or (bit.or (bit.shift fraction 13) (bit.shift exponentF 23)) (bit.shift sign 31)
	(bit.intasfloat outputasfloat*2)
)
fn Read8 file signed:true = (
	number = ReadByte file (if signed then #signed else #unsigned)
	if signed then number /=  127. else number /= 255.
	number
)
fn Read8N file = ((ReadByte file #unsigned)-127)*0.0078125
fn Read8Ns file = (255-(ReadByte file #unsigned))*0.003921568627451

fn RHMRuntime fl bf =(
	items = #()
	chunk = RHMCheckLine fl
	while not (eof fl) and chunk != undefined do (
		case chunk[1] of (
			1: (append items (RHMGeometry fl:fl);chunk = items[items.count].nextline)
			2: (append items (RHMMesh fl:fl);chunk = items[items.count].nextline)
			3: (append items (RHMChunk fl:fl);chunk = items[items.count].nextline)
			default: (format "Unknown Block: %\n" chunk[2];chunk = RHMCheckLine fl)
		)
	)
	close fl
	--print items

	geohdr; for c in items where (classof c) == RHMGeometry while geohdr == undefined do geohdr = c
	if geohdr == undefined then format "Cannot find Geometry tag"
	else (
		
		lstchnk = undefined
		for ch in items where (classof ch) == RHMChunk do (
			if lstchnk != undefined then (
				lstchnk.vertexcount = ch.BaseIndex-lstchnk.BaseIndex
				ch.vertexcount = geohdr.Vertices-ch.BaseIndex
			)	
			lstchnk = ch
		)
		for ch in items where (classof ch) == RHMChunk do (
			if ch.vertexcount == undefined then ch.vertexcount = geohdr.vertices
			for v = 1 to ch.vertexcount do (
				append ch.varray ([readfloat bf,readfloat bf,readfloat bf]*30)
				readshort bf;readbyte bf
				alp1 = Read8ns bf 
				readshort bf;readbyte bf
				alp2 = Read8ns bf 
				append ch.narray [Read8N bf, Read8N bf, Read8N bf]
				alp3 = Read8ns bf 
				append ch.clarray ([alp1,alp2,alp3]+0.4)
				append ch.uvarray [readHalfFloat bf,1-(readHalfFloat bf),0]
				readlong bf	
			)
		)
		materials = #()
		for m = 1 to geohdr.meshes do (
			fndcnt = 0
			fnditm
			for _m in items where (classof _m) == RHMMesh while fndcnt < m do (fnditm=_m;fndcnt += 1)
			fndcnt = 0
			fndchnks = #()
			for ch in items where (classof ch) == RHMChunk while fndcnt < (fnditm.ChunkStart+fnditm.ChunkCount) do (
				if fndcnt >= fnditm.ChunkStart then append fndchnks ch
				fndcnt += 1
			)
			curchunk = 0
			for ch in fndchnks do (
				farray = #()
				for f = 1 to ch.PrimCount do append farray ([readshort bf #unsigned, readshort bf #unsigned, readshort bf #unsigned]+1)
				msh = mesh vertices:ch.varray faces:farray name:(fnditm.name+"["+curchunk as string +"]")
				for j = 1 to ch.narray.count do setNormal msh j ch.narray[j]
				meshop.setNumMaps msh 3
				meshop.setMapSupport msh -2 true
				meshop.buildMapFaces msh 0
				meshop.buildMapFaces msh 1
				meshop.buildMapFaces msh -2
				meshop.setNumMapVerts msh 0 ch.varray.count
				meshop.setNumMapVerts msh 1 ch.uvarray.count
				meshop.setNumMapVerts msh -2 ch.varray.count	
				for k = 1 to ch.uvarray.count do meshop.setMapVert msh 1 k ch.uvarray[k]
				for k = 1 to ch.clarray.count do meshop.setMapVert msh 0 k ch.clarray[k]
				for k = 1 to ch.alarray.count do meshop.setMapVert msh -2 k ch.alarray[k]
				for k = 1 to farray.count do (
					meshop.setMapFace msh -2 k farray[k]
					meshop.setMapFace msh 0 k farray[k]
					meshop.setMapFace msh 1 k farray[k]
				)
				
				if ch.material != undefined then (
					fndmat; for mt in materials where mt.name == ch.material while fndmat == undefined do fndmat = mt					
					if fndmat == undefined then (
						fndmat = standardmaterial name: ch.material
						if ch.DiffuseLayer != undefined then (
							fndmat.diffusemap = Mix map1: (bitmaptexture filename:ch.DiffuseLayer) mask: (vertexcolor())
							if ch.fLayerUVTile != undefined then (fndmat.diffusemap.map1.coords.U_Tiling = ch.fLayerUVTile; fndmat.diffusemap.map1.coords.V_Tiling = ch.fLayerUVTile)
						)
						if ch.Diffuse != undefined then (
							tmptxmap = bitmaptexture filename:ch.Diffuse
							if ch.fBaseUVTile != undefined then (tmptxmap.coords.U_Tiling = ch.fBaseUVTile; tmptxmap.coords.V_Tiling = ch.fBaseUVTile)
							if fndmat.diffusemap == undefined then fndmat.diffusemap = tmptxmap
							else fndmat.diffusemap.map2 = tmptxmap
						)
						
						if ch.SpecularLayer != undefined then fndmat.specularMap = Mix map1: (bitmaptexture filename:ch.SpecularLayer) mask: (vertexcolor())
						if ch.Specular != undefined then if fndmat.specularMap == undefined then fndmat.specularMap = bitmaptexture filename:ch.Specular
						else (fndmat.specularMap.map2 = bitmaptexture filename:ch.Specular)
						
						if ch.NormalLayer != undefined then fndmat.bumpmap = normal_bump normal_map:(Mix map1: (bitmaptexture filename:ch.NormalLayer) mask: (vertexcolor()))
						if ch.Normal != undefined then if fndmat.bumpmap == undefined then fndmat.bumpmap = normal_bump normal_map:(bitmaptexture filename:ch.Normal)
						else (fndmat.bumpmap.normal_map.map2 = bitmaptexture filename:ch.Normal)
							
						if ch.Emissive != undefined then fndmat.selfillummap = bitmaptexture filename:ch.Emissive
						if ch.fSpecularMultiplier != undefined then fndmat.specularLevel = ch.fSpecularMultiplier*10
						if ch.fSpecularPowerMultiplier != undefined then fndmat.Glossiness = ch.fSpecularPowerMultiplier
					--	if ch.fEnvMultiplier != undefined then fndmat.Glossiness = ch.fEnvMultiplier
						if ch.fEmissiveMultiplier != undefined then fndmat.selfIllumMapAmount = ch.fEmissiveMultiplier*100
						append materials fndmat
					)
					msh.material = fndmat
				)
				curchunk +=1
				free farray
			)
		)
	)
	free items
	gc()
)

txfile = getOpenFileName caption:"Open META" types:"META|*.META|All|*.*" historyCategory:"RDHGMETA"
--txfile = @"r:\UE4_DDC\air_conditioner_01/air_conditioner_01_a.meta"
	if txfile != undefined then (
		rhmfile = (getfilenamepath txfile) + (getfilenamefile txfile)+".rhm"
		bf = fopen rhmfile "rb"
		fl = openFile txfile mode:"rt"
		if fl == undefined then MessageBox "Cannot open meta file."
		else if bf == undefined then MessageBox "Cannot open rhm file."
		else (
			RHMRuntime fl bf
			close fl
			fclose bf
		)
	)
-- You want to crash program by uncommenting following line? Cool.
--	filein @"r:\CurUser\Dropbox\PROJEKTY_compiled\OSTATN\MYScripts\roadhog (meta).ms" 